package cmdUser

import (
	"encoding/csv"
	"io"
	"os"
	"strings"

	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/utils"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/caps"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"
	g "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/globals"
	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/utils"
)

type Add struct {
	Name string
}

func (cmd *Add) GetName() string {
	return cmd.Name
}

func (cmd *Add) GetDesc() []string {
	return []string{
		"Creates a user.",
		"Requires USER_CREATE user capability."}
}

func (cmd *Add) PrintUsage() {
	for _, desc := range cmd.GetDesc() {
		u.PrintlnErr(desc)
	}
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("USAGE: " + cmd.GetName() + " email firstname lastname password [capability ...]")
	u.PrintlnErr("       " + cmd.GetName() + " file.csv")
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	usage := `file.csv   5-column CSV file defining each new user:
               column1    email
               column2    firstname
               column3    lastname
               column4    password
               column5    capabilities (space separated)
List of user capabilities:
`
	caps := caps.GetUserCapsNames()
	i := 0
	for _, v := range caps {
		if i == 0 {
			usage += "           "
		}
		usage += v + " "
		i += 1
		if i >= 5 {
			usage = utils.AppendNewLine(usage)
			i = 0
		}
	}
	u.PrintlnErr(usage)
}

func (cmd *Add) Run(client *nc.NexusClient, args []string) int {
	statusCode := 0

	argc := len(args)
	if argc >= 4 {
		// Detected syntax: email firstname lastname pwd [caps]
		p := &params.UserWithPwd{
			Email:     args[0],
			FirstName: args[1],
			LastName:  args[2],
			Pwd:       args[3],
			Caps:      make(caps.Capabilities),
		}
		if argc > 4 {
			for _, cap := range args[4:] {
				p.Caps[cap] = 1
			}
		}

		if err := client.UserCreate(*p); err != nil {
			u.PrintlnErr("Failed creating user " + p.Email + ": " + err.Error())
			statusCode = 1
		} else {
			u.Println("Successfully created user " + p.Email)
		}
	} else if argc == 1 {
		// Detected syntax: file.csv
		csvFile := args[0]
		file, err := os.Open(csvFile)
		if err != nil {
			u.PrintlnErr(err)
			return 1
		}
		defer file.Close()
		reader := csv.NewReader(file)
		reader.Comma = g.CsvFieldSeparator // specify the column separator
		reader.Comment = '#'               // lines starting with this symbol are to be ignored
		line := 0
		for {
			line += 1
			columns, err := reader.Read()
			if err == io.EOF {
				break
			}
			if err != nil {
				u.PrintlnErr("Failed reading " + err.Error())
				continue
			}

			columnCount := len(columns)
			if columnCount < 5 {
				u.PrintlnErr("Failed reading record on line ", line, ": expecting 5 columns")
				statusCode = 1
				continue
			}

			email := columns[0]
			if !utils.IsEmail(email) {
				u.PrintlnErr("Failed reading record on line ", line, ": ", email, " is not a valid email")
				statusCode = 1
				continue
			}

			p := &params.UserWithPwd{
				Email:     email,
				FirstName: columns[1],
				LastName:  columns[2],
				Pwd:       columns[3],
				Caps:      make(map[string]int),
			}
			capsStr := strings.TrimSpace(columns[4])
			if len(capsStr) > 0 {
				caps := strings.Split(capsStr, " ")
				for _, cap := range caps {
					p.Caps[cap] = 1
				}
			}

			if err := client.UserCreate(*p); err != nil {
				u.PrintlnErr("Failed creating user " + p.Email + ": " + err.Error())
				statusCode = 1
			} else {
				u.Println("Successfully created user " + p.Email)
			}
		}
	} else {
		cmd.PrintUsage()
		return 1
	}

	return statusCode
}
