package main

import (
	"bytes"
	_ "embed"
	"errors"
	"os"
	"os/exec"
	"path"
	"time"

	"fyne.io/fyne/v2"
	"fyne.io/fyne/v2/app"
	"fyne.io/fyne/v2/container"
	"fyne.io/fyne/v2/theme"
	"fyne.io/fyne/v2/widget"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/defaults"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/utils"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/version"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/buildversion"
	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
)

const (
	windowTitle    = "nexus-exam"
	connectToVMmsg = "Enter password to connect to your VM:"
)

var (
	//go:embed nexus_exam_user.val
	nexus_exam_user string

	//go:embed nexus_exam_pwd.val
	nexus_exam_pwd string

	client              *nc.NexusClient
	exitFn              func()
	isAuthenticated     bool
	isServerUp          bool
	statusLabel         *widget.Label
	credentialsInputBox *widget.Form
)

func exit(code int) {
	if exitFn != nil {
		exitFn()
	}
	os.Exit(code)
}

func attachVM(parent fyne.Window, pwd string) {
	err := client.VMAttachFromPwd(pwd, true)
	if err != nil {
		errorPopup(parent, "Failed attaching to VM: "+err.Error())
		return
	}
}

func abortWindow(msg string) {
	a := app.New()
	a.Settings().SetTheme(theme.LightTheme())
	win := a.NewWindow(windowTitle)
	win.SetOnClosed(func() {
		win.Close()
		exit(1)
	})
	label := widget.NewLabel("FATAL ERROR: " + msg)
	button := widget.NewButton("Quit", func() {
		win.Close()
		exit(1)
	})
	content := container.NewPadded(container.NewVBox(label, button))
	win.SetContent(content)
	win.ShowAndRun()
}

func errorPopup(win fyne.Window, msg string) {
	var modal *widget.PopUp
	modal = widget.NewModalPopUp(
		container.NewVBox(
			widget.NewLabel("Error: "+msg),
			widget.NewButton("Close", func() { modal.Hide() })),
		win.Canvas())
	modal.Show()
}

func hypervisorCheck() {
	cmd := exec.Command("/usr/bin/systemd-detect-virt")
	var out bytes.Buffer
	cmd.Stdout = &out
	cmd.Run()
	// Returns "none\n" if running on a physical system (no VM/container)
	if out.String() != "none\n" {
		abortWindow("nexus-exam only works on a physical system!")
	}
}

// Recurrently obtains a new JWT token so that the user session doesn't expire.
func refreshToken(parent fyne.Window) {
	var err error

	for {
		if isServerUp {
			if !isAuthenticated {
				u.Println(time.Now(), ": initiate authentication")
				_, err = client.Authenticate(nexus_exam_user, nexus_exam_pwd)
				if err != nil {
					statusLabel.SetText("ERROR: failed to authenticate")
					u.PrintlnErr(time.Now(), ": failed to authenticate: "+err.Error())
					time.Sleep(30 * time.Second)
				} else {
					u.Println(time.Now(), ": successfully authenticated")
					statusLabel.SetText(connectToVMmsg)
					isAuthenticated = true
					credentialsInputBox.Show()
					time.Sleep(60 * time.Second)
				}
			} else {
				u.Println(time.Now(), ": attempt to refresh token")
				client.RefreshToken()
				if err != nil {
					statusLabel.SetText("ERROR: failed to refresh token")
					u.PrintlnErr(time.Now(), ": failed to refresh token: "+err.Error())
					isAuthenticated = false
					credentialsInputBox.Hide()
				} else {
					statusLabel.SetText(connectToVMmsg)
					u.Println(time.Now(), ": successfully refreshed token")
					credentialsInputBox.Show()
					time.Sleep(60 * time.Second)
				}
			}
		} else {
			time.Sleep(1 * time.Second)
		}
	}
}

func checkServerUp(parent fyne.Window) {
	for {
		// Instead of implementing a keepalive service (on both server and client side),
		// we simply use the server version route to know whether the server is up or down.
		_, err := client.GetServerVersion()
		if err != nil {
			u.Println(time.Now(), ": server is down: "+err.Error())
			isServerUp = false
			credentialsInputBox.Hide()
			statusLabel.SetText("ERROR: server cannot be reached or service is down")
		} else {
			isServerUp = true
		}

		time.Sleep(10 * time.Second)
	}
}

func run() int {
	hypervisorCheck()

	var err error
	client, err = nc.New(defaults.NexusServer)
	if err != nil {
		abortWindow(err.Error())
	}

	exitFn = func() { client.Cleanup() }

	// This thread acts as a signal handler for SIGINT or SIGTERM.
	// When one of these signals is received, the temporary certificate file is deleted.
	// Without this "handler", the temporary certificate file wouldn't be deleted.
	go func() {
		u.WaitForSignals()
		exit(1)
	}()

	isServerUp = false
	isAuthenticated = false
	var appname = path.Base(os.Args[0])
	clientVersion := version.Get()
	u.PrintlnErr(appname + " version " + clientVersion.String() + " (commit " + buildversion.GitCommit() + ")")
	u.PrintlnErr("Built on " + buildversion.BuildDate())

	// Checks the client version is compatible with the server's API.
	// if !cmdVersion.CheckServerCompatibility("nexus-exam") {
	//     abortWindow("client version is incompatible with server!")
	// }

	app := app.New()
	app.Settings().SetTheme(theme.LightTheme())
	win := app.NewWindow(windowTitle)

	pwdEntry := widget.NewEntry()
	pwdEntry.Password = true

	pwdEntry.Validator = func(val string) error {
		if len(val) < 8 {
			return errors.New("Please enter at least 8 characters")
		}
		return nil
	}

	statusLabel = widget.NewLabel("\n\n\n")
	statusLabel.Show()

	credentialsInputBox = &widget.Form{
		Items: []*widget.FormItem{
			{Text: "Password", Widget: pwdEntry},
		},
		OnSubmit: func() {
			attachVM(win, pwdEntry.Text)
		},
		SubmitText: "Connect",
	}

	win.SetContent(container.NewPadded(container.NewVBox(statusLabel, credentialsInputBox)))
	win.Resize(fyne.NewSize(600, 160))
	credentialsInputBox.Hide()

	go checkServerUp(win)
	go refreshToken(win)

	win.ShowAndRun()
	return 0
}

func main() {
	exit(run())
}
