package main

import (
	"errors"
	"fmt"
	"io"
	"os"
	"path"
	"strings"
	"syscall"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmd"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmdMisc"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmdTemplate"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmdToken"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmdUser"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmdVM"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmdVersion"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/defaults"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/utils"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/version"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/buildversion"
	g "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/globals"
	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"github.com/peterh/liner"
	"golang.org/x/term"
)

var cmdList = []cmd.Command{
	&cmdMisc.HelpHeader{Name: "!═════╡ GENERAL commands ╞═════════════════════════════════════════════════════════════════"},
	// Commands in this block are specific to nexush:
	&cmdMisc.Ls{Name: "ls"},
	// End of nexush specific commands

	&cmdToken.Refresh{Name: "refresh"},
	&cmdVersion.Version{Name: "version"},

	&cmdMisc.HelpHeader{Name: "!═════╡ USER commands ╞═════════════════════════════════════════════════════════════════"},
	&cmdUser.Whoami{Name: "whoami"},
	&cmdUser.UpdatePwd{Name: "passwd"},
	&cmdUser.List{Name: "userlist"},
	&cmdUser.Add{Name: "usercreate"},
	&cmdUser.Del{Name: "userdel"},
	&cmdUser.ResetPwd{Name: "userresetpwd"},
	&cmdUser.SetCaps{Name: "usersetcaps"},
	&cmdUser.Unlock{Name: "userunlock"},

	&cmdMisc.HelpHeader{Name: "!═════╡ TEMPLATE commands ╞═════════════════════════════════════════════════════════════════"},
	&cmdTemplate.Create{Name: "tplcreate"},
	&cmdTemplate.Del{Name: "tpldel"},
	&cmdTemplate.Edit{Name: "tpledit"},
	&cmdTemplate.ExportDisk{Name: "tplexportdisk"},
	&cmdTemplate.List{Name: "tpllist"},
	// &cmdTemplate.ListSingle{Name: "tpllistsingle"}, // for testing the route only

	&cmdMisc.HelpHeader{Name: "!═════╡ VM commands ╞═════════════════════════════════════════════════════════════════"},
	&cmdVM.AddAccess{Name: "vmaddaccess"},
	&cmdVM.Attach{Name: "vmattach"},
	// &cmdVM.AttachSingle{Name: "vmattachsingle"},   // for testing the route only
	// &cmdVM.AttachFromPwd{Name: "vmattachfrompwd"}, // for testing the route only
	&cmdVM.Create{Name: "vmcreate"},
	&cmdVM.Creds2pdf{Name: "vmcreds2pdf"},
	&cmdVM.Creds2csv{Name: "vmcreds2csv"},
	&cmdVM.Del{Name: "vmdel"},
	&cmdVM.DelAccess{Name: "vmdelaccess"},
	&cmdVM.Edit{Name: "vmedit"},
	&cmdVM.ExportDir{Name: "vmexportdir"},
	&cmdVM.ImportDir{Name: "vmimportdir"},
	&cmdVM.Stop{Name: "vmkill"},
	&cmdVM.List{Name: "vmlist"},
	// &cmdVM.ListSingle{Name: "vmlistsingle"},   // for testing the route only
	&cmdVM.Reboot{Name: "vmreboot"},
	&cmdVM.Shutdown{Name: "vmshutdown"},
	&cmdVM.Start{Name: "vmstart"},
	&cmdVM.StartAttach{Name: "vmstartattach"},
	&cmdVM.StartWithCreds{Name: "vmstartwithcreds"},
}

var exitFn func()
var prompt *liner.State = nil
var savedTermState *term.State = nil

func exit(code int) {
	if exitFn != nil {
		exitFn()
	}
	os.Exit(code)
}

func run() int {
	var appname = path.Base(os.Args[0])
	clientVersion := version.Get()
	u.PrintlnErr(appname + " version " + clientVersion.String() + " (commit " + buildversion.GitCommit() + ")")
	u.PrintlnErr("Built on " + buildversion.BuildDate())

	if len(os.Args) < 2 {
		u.PrintlnErr("USAGE: " + appname + " EMAIL")
		u.PrintlnErr("EMAIL is the user with which to log in.")
		return 1
	}

	serverEnvVar, found := os.LookupEnv(g.ENV_NEXUS_SERVER)
	if !found {
		serverEnvVar = defaults.NexusServer
		// u.PrintlnErr("Environment variable \""+g.ENV_NEXUS_SERVER+"\" must be set!")
		// u.PrintlnErr("It defines the nexus server to connect to along the port number.")
		// u.PrintlnErr("Example: export "+g.ENV_NEXUS_SERVER+"=192.168.1.42:1077")
		// return 1
	}

	savedTermState, _ = term.GetState(int(os.Stdin.Fd()))

	client, err := nc.New(serverEnvVar)
	if err != nil {
		u.PrintlnErr("Error: " + err.Error())
		return 1
	}

	// Checks the client version is compatible with the server's API.
	if !cmdVersion.CheckServerCompatibility(client, appname) {
		return 1
	}

	// Obtains password from cmd line.
	email := os.Args[1]
	fmt.Printf("%s", email+"'s password: ")
	bytePwd, err := term.ReadPassword(int(syscall.Stdin))
	if err != nil {
		u.PrintlnErr(err)
		return 1
	}
	pwd := string(bytePwd)

	// Authenticate the user (internally obtains a JWT token).
	_, err = client.Authenticate(email, pwd)
	if err != nil {
		u.PrintlnErr("Error: " + err.Error())
		return 1
	}

	exitFn = func() { client.Cleanup() }

	// This thread acts as a signal handler for SIGINT or SIGTERM.
	// When one of these signals is received, the temporary certificate file is deleted.
	// Without this "handler", the temporary certificate file wouldn't be deleted.
	go func() {
		u.WaitForSignals()
		restoreTerm()
		exit(1)
	}()

	shell(client)
	return 0
}

func shell(client *nc.NexusClient) {
	u.Println(`Welcome to nexush, the nexus shell.
Type: "help" for help on commands
      "quit" or "exit" to quit nexush`)

	prompt = liner.NewLiner()
	prompt.SetCtrlCAborts(true)

	prompt.SetCompleter(func(line string) (commands []string) {
		for _, cmd := range cmdList {
			c := cmd.GetName()
			if strings.HasPrefix(c, strings.ToLower(line)) {
				commands = append(commands, c)
			}
		}
		return
	})

	quit := false
	for !quit {
		var line string
		if myline, err := prompt.Prompt("nexush> "); err == nil {
			line = myline
			prompt.AppendHistory(myline)
		} else if err == liner.ErrPromptAborted {
			u.Println("Aborted")
			continue
		} else if err == io.EOF {
			u.Println("")
			quit = true
			continue
		} else {
			u.PrintlnErr("Error reading line: ", err)
			continue
		}

		line = strings.TrimSpace(line)
		args, err := extractArgs(line)
		if err != nil {
			u.PrintlnErr("invalid command line: " + err.Error())
			continue
		}
		if args == nil {
			continue
		}
		if len(args) == 0 {
			continue
		}
		typedCmd := args[0]

		switch {
		case typedCmd == "quit" || typedCmd == "exit":
			quit = true
		case typedCmd == "help":
			cmd.Help(cmdList)
		case line == "":
			continue
		default:
			found, cmd := cmd.Match(typedCmd, cmdList)
			if found {
				if typedCmd == "refresh" {
					err := client.RefreshToken()
					if err != nil {
						u.PrintlnErr("refresh error: " + err.Error())
						break
					}
					u.Println("Token refreshed.")
					break
				} else {
					cmd.Run(client, args[1:])
					break
				}
			} else {
				u.PrintlnErr(typedCmd + ": unknown command")
			}
		}
	}

	restoreTerm()
}

func restoreTerm() {
	if prompt != nil {
		prompt.Close()
	}
	if savedTermState != nil {
		term.Restore(int(os.Stdin.Fd()), savedTermState)
	}
}

// Extract arguments from a command line, taking in account args between quotes ("").
// Returns arguments as a slice of strings.
func extractArgs(line string) ([]string, error) {
	toggleReplace := false
	var sb strings.Builder
	quoteCount := 0
	for _, c := range line {
		if c == ' ' {
			if toggleReplace {
				// IMPORTANT:
				// The rune below is used to replace spaces within quotes in order
				// to split the string on spaces as normal. Then, after the split,
				// we simply replace the rune with spaces.
				// Consequently the input string MUST NOT contain this specific rune!
				sb.WriteRune('妖')
			} else {
				sb.WriteRune(c)
			}
		} else if c == '"' {
			toggleReplace = !toggleReplace
			quoteCount += 1
		} else {
			sb.WriteRune(c)
		}
	}

	if quoteCount%2 == 1 {
		return nil, errors.New("missing closing quote")
	}

	args := strings.Split(sb.String(), " ")
	newArgs := []string{}

	for _, arg := range args {
		r := strings.NewReplacer("妖", " ")
		str := r.Replace(arg)
		str = strings.TrimSpace(str)
		if len(str) > 0 {
			newArgs = append(newArgs, str)
		}
	}

	return newArgs, nil
}

func main() {
	exit(run())
}
